#!/usr/bin/env python

# Example call: python -u notes/cdhit_uniq_stats.py 10 "/scratch/users/vgalata/GDB/results/analysis/cdhit/" | tee notes/cdhit_uniq_stats_GDB.tsv

import os
import re
import sys
import glob
import pandas
from multiprocessing import Pool

def parse_faa(faa_file):
    df = []
    rtype1 = os.path.basename(faa_file).split("__")[0].split("_")[0]
    tool1  = "_".join(os.path.basename(faa_file).split("__")[0].split("_")[1:])
    with open(faa_file, "r") as ifile:
        for line in ifile:
            if not re.match("^>", line):
                continue
            line = re.sub("^>", "", line)
            sid, sstart, send, sstrand, sinfo = line.split(" # ")
            srtype, stool, sid = sid.split("__")
            scontigid = "_".join(sid.split("_")[:-1])
            sgeneid = sid.split("_")[-1]
            spartial = sinfo.split(";")[1]
            slen = int(send) - int(sstart) + 1
            sdict = {
                "rtype1": rtype1,
                "tool1": tool1,
                "rtype2": srtype,
                "tool2": stool,
                "contig_id": scontigid,
                "gene_id": sgeneid,
                "start": int(sstart),
                "end": int(send),
                "strand": int(sstrand),
                "partial": spartial,
                "length": slen
            }
            df.append(sdict.copy())
    df = pandas.DataFrame(df)
    return df

def find_clusters(df):
    assert len(set(df.rtype1)) == 1 and len(set(df.tool1)) == 1 and len(set(df.rtype2)) == 1 and len(set(df.tool2)) == 1
    df = df.sort_values(by=["contig_id", "gene_id"], axis=0, inplace=False)
    df = df.assign(cluster_label=0)
    df = df.assign(cluster=False)
    last_i = ""
    last_cid = ""
    last_gid = -1
    cluster  = 0
    for i in df.index:
        i_cid = df.loc[i, "contig_id"]
        i_gid = int(df.loc[i, "gene_id"])
        # same contig, consecutive genes
        if last_cid == i_cid and last_gid + 1 == i_gid:
            df.loc[last_i, "cluster"] = True
            df.loc[i, "cluster"]      = True
        else:
            cluster += 1
        df.loc[i, "cluster_label"] = cluster
        last_i = i
        last_cid = i_cid
        last_gid = i_gid
    return df

def print_stats(df):
    # stats
    gene_total_count            = df.shape[0]
    gene_ave_length             = df["length"].mean()
    cluster_total_count         = len(set(df.loc[df.cluster,"cluster_label"]))
    cluster_gene_total_count    = sum(df.cluster)
    cluster_gene_ave_count      = df.loc[df.cluster,:].groupby("cluster_label")["gene_id"].agg(["size"]).mean().values[0]
    cluster_gene_ave_length     = df.loc[df.cluster,:]["length"].mean()
    # print
    # df_lb = (df.rtype1[0], df.tool1[0], df.rtype2[0], df.tool2[0])
    # print("\
    # genes in {} but NOT in {}\n\
    # \t{} genes w/ ave. length of {:.2f}bp\n\
    # \t{} clusters w/ {} ({:.2f}%) genes in total\n\
    # \tave. number of genes in clusters: {:.2f}\n\
    # \tave. length of genes in clusters: {:.2f}bp\n\
    # ".format(
    #     df_lb[2:], df_lb[0:2],
    #     gene_total_count, gene_ave_length,
    #     cluster_total_count, cluster_gene_total_count, 100 * cluster_gene_total_count / gene_total_count,
    #     cluster_gene_ave_count,
    #     cluster_gene_ave_length
    # ))
    print("{rt1}\t{to1}\t{rt2}\t{to2}\t{gtotal}\t{gavelen:.2f}\t{ctotal}\t{cgtotal}\t{cgpct:.2f}\t{cgavecount:.2f}\t{cgavelen:.2f}".format(
        rt1=df.rtype1[0], to1=df.tool1[0], rt2=df.rtype2[0], to2=df.tool2[0],
        gtotal=gene_total_count, gavelen=gene_ave_length,
        ctotal=cluster_total_count,
        cgtotal=cluster_gene_total_count, cgpct=100*cluster_gene_total_count/gene_total_count,
        cgavecount=cluster_gene_ave_count,
        cgavelen=cluster_gene_ave_length
    ))
    return

def proc_input(faa_file):
    df = parse_faa(faa_file)
    df = find_clusters(df)
    print_stats(df)

if __name__ == "__main__":
    # files to process
    faa_files = sorted(glob.glob(os.path.join(sys.argv[2], "*.faa")))
    # header
    print("\t".join([
        "rtype1", "tool1", "rtype2", "tool2",
        "gene_total_count", "gene_ave_length",
        "cluster_total_count", "cluster_gene_total_count", "cluster_gene_total_pct",
        "cluster_gene_ave_count", "cluster_gene_ave_length"
    ]))
    # stat.s
    with Pool(int(sys.argv[1])) as p:
        p.map(proc_input, faa_files)
    