# Assembly

##################################################
# Long reads

rule assembly_lr_flye:
    input:
        os.path.join(RESULTS_DIR, "preproc/metag/lr/lr.proc.fastq.gz")
    output:
        os.path.join(RESULTS_DIR, "assembly/lr/flye/ASSEMBLY.fasta")
    threads:
        config["flye"]["threads"]
    log:
        "logs/assembly.lr.flye.log"
    conda:
        os.path.join(ENV_DIR, "flye.yaml")
    message:
        "Assembly: long reads: Flye"
    shell:
        "(date && "
        "flye --nano-raw {input} --meta --out-dir $(dirname {output}) --threads {threads} && "
        "cd $(dirname {output}) && ln -s assembly.fasta $(basename {output}) && "
        "date) &> {log}"

rule assembly_lr_canu:
    input:
        os.path.join(RESULTS_DIR, "preproc/metag/lr/lr.proc.fastq.gz")
    output:
        os.path.join(RESULTS_DIR, "assembly/lr/canu/ASSEMBLY.fasta")
    log:
        "logs/assembly.lr.canu.log"
    threads:
        config["canu"]["threads"]
    conda:
        os.path.join(ENV_DIR, "canu.yaml")
    message:
        "Assembly: long reads: canu"
    shell:
        "(date && "
        "canu -nanopore-raw {input} -p \"assembly\" -d $(dirname {output}) genomeSize={config[canu][genome_size]} "
        "useGrid=false minThreads={threads} maxThreads={threads} minInputCoverage=8 stopOnLowCoverage=8 && "
        "cd $(dirname {output}) && ln -s assembly.contigs.fasta $(basename {output}) && "
        "date) &> {log}"

# localrules: assembly_lr_link

# rule assembly_lr_link:
#     input:
#         os.path.join(RESULTS_DIR, "assembly/lr/{tool}/polishing_raconLR/assembly.fasta")
#     output:
#         os.path.join(RESULTS_DIR, "assembly/lr/{tool}/ASSEMBLY.fasta")
#     wildcard_constraints:
#         tool="|".join([tool for tool in config["assemblers"]["lr"] if not re.search("_sr$", tool)]),
#     shell:
#         "ipath=$(realpath {input}) && cd $(dirname {output}) && ln -sf $(realpath --relative-to=\".\" \"${{ipath}}\") $(basename {output})"

# localrules: assembly_lr_polished_sr

# rule assembly_lr_polished_sr:
#     input: 
#         os.path.join(RESULTS_DIR, "assembly/lr/{tool}/polishing_raconLR_raconSR/assembly.fasta")
#     output: 
#         os.path.join(RESULTS_DIR, "assembly/lr/{tool}/ASSEMBLY.fasta")
#     wildcard_constraints:
#         tool="|".join([tool for tool in config["assemblers"]["lr"] if re.search("_sr$", tool)]),
#     shell:
#         "ipath=$(realpath {input}) && cd $(dirname {output}) && ln -sf $(realpath --relative-to=\".\" \"${{ipath}}\") $(basename {output})"

##################################################
# Short reads

rule assembly_sr_megahit:
    input: 
        r1=os.path.join(RESULTS_DIR, "preproc/metag/sr/R1.proc.fastq.gz"),
        r2=os.path.join(RESULTS_DIR, "preproc/metag/sr/R2.proc.fastq.gz")
    output:
        os.path.join(RESULTS_DIR, "assembly/sr/megahit/ASSEMBLY.fasta")
    log:
        "logs/assembly.sr.megahit.log",
    threads:
        config["megahit"]["threads"]
    conda:
        os.path.join(ENV_DIR, "megahit.yaml")
    message:
        "Assembly: short reads: MEGAHIT"
    shell:
        "(date && megahit -1 {input.r1} -2 {input.r2} -t {threads} -o $(dirname {output})/tmp && "
        "cd $(dirname {output}) && "
        "rsync -avP tmp/ . && "
        "ln -sf final.contigs.fa $(basename {output}) && "
        "rm -rf tmp/ && "
        "date) &> {log}"

rule assembly_sr_metaspades:
    input:
        r1=os.path.join(RESULTS_DIR, "preproc/metag/sr/R1.proc.fastq.gz"),
        r2=os.path.join(RESULTS_DIR, "preproc/metag/sr/R2.proc.fastq.gz")
    output:
        os.path.join(RESULTS_DIR, "assembly/sr/metaspades/ASSEMBLY.fasta")
    log:
        "logs/assembly.sr.metaspades.log"
    threads:
        config["metaspades"]["threads"]
    conda:
        os.path.join(ENV_DIR, "spades.yaml")
    message:
        "Assembly: short reads: MetaSPAdes"
    shell:
        "(date && "
        "metaspades.py -k 21,33,55,77 -t {threads} -1 {input.r1} -2 {input.r2} -o $(dirname {output}) && "
        "cd $(dirname {output}) && ln -sf contigs.fasta $(basename {output}) && "
        "date) &> {log}"

##################################################
# Hybrid

rule assembly_hy_metaspades:
    input: 
        lr=os.path.join(RESULTS_DIR, "preproc/metag/lr/lr.proc.fastq.gz"),
        r1=os.path.join(RESULTS_DIR, "preproc/metag/sr/R1.proc.fastq.gz"),
        r2=os.path.join(RESULTS_DIR, "preproc/metag/sr/R2.proc.fastq.gz")
    output: 
        os.path.join(RESULTS_DIR, "assembly/hy/metaspadeshybrid/ASSEMBLY.fasta")
    log:
        "logs/assembly.hy.metaspadeshybrid.log"
    threads:
        config["metaspades"]["threads"]
    conda:
        os.path.join(ENV_DIR, "spades.yaml")
    message:
        "Assembly: hybrid: MetaSPAdes"
    shell:
        "(date && "
        "spades.py --meta -k 21,33,55,77 -t {threads} -1 {input.r1} -2 {input.r2} --nanopore {input.lr} -o $(dirname {output}) && "
        "cd $(dirname {output}) && ln -sf contigs.fasta $(basename {output}) && "
        "date) &> {log}"

rule assembly_hy_operams:
    input: 
        lr=os.path.join(RESULTS_DIR, "preproc/metag/lr/lr.proc.fastq.gz"),
        r1=os.path.join(RESULTS_DIR, "preproc/metag/sr/R1.proc.fastq.gz"),
        r2=os.path.join(RESULTS_DIR, "preproc/metag/sr/R2.proc.fastq.gz"),
        asm=os.path.join(RESULTS_DIR, "assembly/sr/{asm}/ASSEMBLY.fasta"),
        bin=os.path.join(MOD_DIR, "operams/OPERA-MS.pl"),
        install=os.path.join(MOD_DIR, "operams/installation.done")
    output:
        asm=os.path.join(RESULTS_DIR, "assembly/hy/operams{asm}/ASSEMBLY.fasta"),
        lr=temp(os.path.join(RESULTS_DIR, "assembly/hy/operams{asm}/lr.fastq"))
    log:
        "logs/assembly.hy.operams.{asm}.log"
    wildcard_constraints:
        asm="|".join(config["assemblers"]["sr"])
    threads:
        config["operams"]["threads"]
    conda:
        os.path.join(ENV_DIR, "operams.yaml")
    message:
        "Assembly: hybrid: OPERA-MS using contigs from {wildcards.asm}"
    shell:
       "(date && "
       "zcat {input.lr} > {output.lr} && "
       "perl {input.bin} --short-read1 {input.r1} --short-read2 {input.r2} --long-read {output.lr} --contig-file {input.asm} "
       "--out-dir $(dirname {output.asm}) --long-read-mapper minimap2 --num-processors {threads} && "
       "cd $(dirname {output.asm}) && ln -sf contigs.fasta $(basename {output.asm}) && "
       "date) &> {log}"

# localrules: assembly_hy_polished_sr

# rule assembly_hy_polished_sr:
#     input: 
#         os.path.join(RESULTS_DIR, "assembly/hy/{tool}_sr/polishing_raconSR/assembly.fasta")
#     output: 
#         os.path.join(RESULTS_DIR, "assembly/hy/{tool}_sr/ASSEMBLY.fasta")
#     wildcard_constraints:
#         tool="|".join(config["assemblers"]["hy"]),
#     shell:
#         "ipath=$(realpath {input}) && cd $(dirname {output}) && ln -sf $(realpath --relative-to=\".\" \"${{ipath}}\") $(basename {output})"

##################################################
# Polishing w/ metaT

# rule polishing_metat_racon:
#     input:
#         fastq=os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}metatracon/reads.fastq"),
#         asm=os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.fasta"),
#         sam=os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}metatracon/mapping.sam")
#     output:
#         fasta=os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}metatracon/ASSEMBLY.fasta"),
#         # NOTE: Tmp files for input data to pass file extension checks
#         asm=temp(os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}metatracon/input.fasta"))
#     log:
#         "logs/racon.metat.{rtype}.{tool}.log"
#     wildcard_constraints:
#         rtype="|".join(["sr", "hy"]),
#         tool="|".join(["metaspades", "megahit", "metaspadeshybrid", "operams"])
#     threads:
#         config["racon"]["threads"]
#     conda:
#         os.path.join(ENV_DIR, "racon.yaml")
#     message:
#         "Assembly: polishing w/ Racon and metaT"
#     shell:
#         "(date && "
#         "cp {input.asm} {output.asm} && "
#         "racon -t {threads} --include-unpolished {input.fastq} {input.sam} {output.asm} > {output.fasta} && "
#         "date) &> {log}"

##################################################
# Polishing

# localrules: polishing_input_sr

# rule polishing_input_sr:
#     input:
#         os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.fasta")
#     output:
#         os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}_sr/assembly.fasta")
#     wildcard_constraints:
#         rtype="|".join(READ_TYPES),
#         tool="|".join([tool for tool in ASSEMBLERS if not re.search("_sr$", tool)])
#     shell:
#         "cp {input} {output}"

# localrules: polishing_input

# rule polishing_input:
#     input:
#         os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/assembly.fasta")
#     output:
#         os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/polishing/assembly.fasta")
#     wildcard_constraints:
#         rtype="|".join(READ_TYPES),
#         tool="|".join([tool for tool in ASSEMBLERS if not re.search("_sr$", tool)])
#     shell:
#         "cd $(dirname {output}) && ln -sf ../$(basename {input}) $(basename {output})"

# rule polishing_racon_lr:
#     input:
#         lr=os.path.join(RESULTS_DIR, "preproc/metag/lr/lr.proc.fastq.gz"),
#         asm=os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/{ptype}/assembly.fasta"),
#     output:
#         fasta=os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/{ptype}_raconLR/assembly.fasta"),
#         bam=os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/{ptype}_raconLR/input.bam"),
#         sam=temp(os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/{ptype}_raconLR/input.sam")),
#         asm=temp(os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/{ptype}_raconLR/input.fasta"))
#     log:
#         "logs/polishing.{ptype}_raconLR.{rtype}.{tool}.log"
#     wildcard_constraints:
#         rtype="|".join(READ_TYPES),
#         tool="|".join(ASSEMBLERS),
#         ptype="polishing.*"
#     threads:
#         config["racon"]["threads"]
#     params:
#         idx_prefix=lambda wildcards, output, input: os.path.splitext(input.asm)[0],
#         bam_prefix=lambda wildcards, output: os.path.splitext(output.bam)[0],
#         chunk_size=config["samtools"]["sort"]["chunk_size_bigmem"]
#     conda:
#         os.path.join(ENV_DIR, "racon.yaml")
#     message:
#         "Assembly polishing w/ Racon and LR"
#     shell:
#         "(date && "
#         # index for mapping
#         "bwa index {input.asm} -p {params.idx_prefix} && "
#         # mapping
#         "bwa mem -x ont2d -t {threads} {params.idx_prefix} {input.lr} | "
#         "samtools view -@ {threads} -SbT {input.asm} | "
#         "samtools sort -@ {threads} -m {params.chunk_size} -T {params.bam_prefix} > {output.bam} && "
#         "samtools view -@ {threads} -h -o {output.sam} {output.bam} && "
#         # copy assembly
#         "cp {input.asm} {output.asm} && "
#         # polish
#         "racon --include-unpolished -t {threads} {input.lr} {output.sam} {output.asm} > {output.fasta} && "
#         "date) &> {log}"

# rule polishing_racon_sr:
#     input:
#         r1=os.path.join(RESULTS_DIR, "preproc/metag/sr/R1.proc.fastq.gz"),
#         r2=os.path.join(RESULTS_DIR, "preproc/metag/sr/R2.proc.fastq.gz"),
#         asm=os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/assembly.fasta"),
#     output:
#         fasta=os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/{ptype}_raconSR/assembly.fasta"),
#         bam=os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/{ptype}_raconSR/input.bam"),
#         sam=temp(os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/{ptype}_raconSR/input.sam")),
#         asm=temp(os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/{ptype}_raconSR/input.fasta")),
#         fastq=temp(os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/{ptype}_raconSR/input.fastq")),
#     log:
#         "logs/polishing.{ptype}_raconSR.{rtype}.{tool}.log"
#     wildcard_constraints:
#         rtype="|".join(READ_TYPES),
#         tool="|".join(ASSEMBLERS),
#         ptype="polishing.*"
#     threads:
#         config["racon"]["threads"]
#     params:
#         idx_prefix=lambda wildcards, output, input: os.path.splitext(input.asm)[0],
#         bam_prefix=lambda wildcards, output: os.path.splitext(output.bam)[0],
#         chunk_size=config["samtools"]["sort"]["chunk_size_bigmem"]
#     conda:
#         os.path.join(ENV_DIR, "racon.yaml")
#     message:
#         "Assembly polishing w/ Racon and LR"
#     shell:
#         "(date && "
#         # index for mapping
#         "bwa index {input.asm} -p {params.idx_prefix} && "
#         # proc reads
#         "zcat {input.r1} | awk '{{print (NR%4 == 1) ? \"@1_\" ++i : $0}}' > {output.fastq} && "
#         "zcat {input.r2} | awk '{{print (NR%4 == 1) ? \"@2_\" ++i : $0}}' >> {output.fastq} && "
#         # mapping
#         "bwa mem -t {threads} {params.idx_prefix} {output.fastq} | "
#         "samtools view -@ {threads} -SbT {input.asm} | "
#         "samtools sort -@ {threads} -m {params.chunk_size} -T {params.bam_prefix} > {output.bam} && "
#         "samtools view -@ {threads} -h -o {output.sam} {output.bam} && "
#         # copy assembly
#         "cp {input.asm} {output.asm} && "
#         # polish
#         "racon --include-unpolished -t {threads} {output.fastq} {output.sam} {output.asm} > {output.fasta} && "
#         "date) &> {log}"

# rule polishing_lr_medaka:
#     input: 
#         lr=os.path.join(RESULTS_DIR, "preproc/metag/lr/lr.proc.fastq.gz"),
#         asm=os.path.join(RESULTS_DIR, "assembly/lr/{tool}/polishing/racon/assembly.fasta")
#     output:
#         os.path.join(RESULTS_DIR, "assembly/lr/{tool}/polishing/racon_medaka/consensus.fasta")
#     log:
#         "logs/racon.medaka.{tool}.log"
#     wildcard_constraints:
#         tool="|".join(config["assemblers"]["lr"])
#     threads:
#         config["medaka"]["threads"]
#     conda:
#         os.path.join(ENV_DIR, "medaka.yaml")
#     message:
#         "Assembly: long reads: polishing w/ Medaka"
#     shell:
#         "(date && "
#         "medaka_consensus -i {input.lr} -d {input.asm} -o $(dirname {output}) -t {threads} -m {config[medaka][model]} && "
#         # "cd $(dirname {output}) && ln -sf consensus.fasta $(basename {output}) && " # NOTE: flye already creates this file
#         "date) &> {log}"

##################################################
# Contig filtering

# rule filter_contigs:
#     input:
#         asm=os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.fasta"),
#         cov=os.path.join(RESULTS_DIR, "mapping/metag/{rtype}/{tool}/ASSEMBLY.{rtype}.cov.ave.txt")
#     output:
#         os.path.join(RESULTS_DIR, "assembly/{rtype}/{tool}/ASSEMBLY.FILTERED.fasta")
#     params:
#         minlen=1000,
#         mincov=5
#     conda:
#         os.path.join(ENV_DIR, "rgi.yaml") # NOTE: need an env. containing Biopython and pandas
#     message:
#         "Assembly: filter contigs"
#     script:
#         os.path.join(SRC_DIR, "filter_contigs.py")